__author__ = 'Georgios Rizos (georgerizos@iti.gr)'

import copy
import networkx as nx
import community
import numpy as np
import scipy.sparse as sparse
import scipy.sparse.linalg as spla

from reveal_graph_embedding.embedding.laplacian import get_normalized_laplacian


def mroc(adjacency_matrix, alpha):
    """
    Extracts hierarchical community features using the MROC method.

    Introduced in: Wang, X., Tang, L., Liu, H., & Wang, L. (2013).
                   Learning with multi-resolution overlapping communities.
                   Knowledge and information systems, 36(2), 517-535.

    Inputs:  - A in R^(nxn): Adjacency matrix of an undirected network represented as a SciPy Sparse COOrdinate matrix.
             - alpha: A maximum community size stopping threshold.

    Outputs: - X in R^(nxC_n): The latent space embedding represented as a SciPy Sparse COOrdinate matrix.
    """
    # Find number of nodes
    number_of_nodes = adjacency_matrix.shape[0]

    ####################################################################################################################
    # Base community calculation
    ####################################################################################################################
    # Initialize empty lists
    base_list = list()
    base_row = list()
    base_col = list()

    # Save function handles for speed
    append_base_list = base_list.append
    append_base_row = base_row.append
    append_base_col = base_col.append

    # Find base communities
    adjacency_matrix = adjacency_matrix.tocsc()
    number_of_base_communities = 0
    for i in range(number_of_nodes):
        # Calculate base community
        base_community = set(adjacency_matrix.getcol(i).indices)
        base_community.add(i)
        flag = True
        for c in base_list:
            if c == base_community:
                flag = False
                break
        if flag:
            append_base_list(base_community)

            for n in base_community:
                append_base_row(n)
                append_base_col(number_of_base_communities)

            number_of_base_communities += 1

    # Form sparse matrices
    base_row = np.array(base_row)
    base_col = np.array(base_col)
    base_data = np.ones(base_row.size, dtype=np.float64)
    features = sparse.coo_matrix((base_data, (base_row, base_col)),
                                 shape=(number_of_nodes, number_of_base_communities))

    features = features.tocsr()
    base_community_number = features.shape[1]

    print('Base communities calculated.')

    reverse_index_csr = copy.copy(features)
    reverse_index_csc = reverse_index_csr.tocsc()
    reverse_index_csr = reverse_index_csr.tocsr()

    reverse_index_rows = np.ndarray(number_of_nodes, dtype=np.ndarray)
    reverse_index_cols = np.ndarray(number_of_nodes, dtype=np.ndarray)
    for n in range(number_of_nodes):
        reverse_index_row = reverse_index_csr.getrow(n)
        reverse_index_rows[n] = reverse_index_row.indices

        if n < base_community_number:
            reverse_index_col = reverse_index_csc.getcol(n)
            reverse_index_cols[n] = reverse_index_col.indices

    flag = True

    print('Start merge iterations.')

    iteration = 0

    while flag:
        level_row = list()
        level_col = list()

        append_level_row = level_row.append
        append_level_col = level_col.append

        unavailable_communities = -1*np.ones(reverse_index_csc.shape[1])
        unavailable_communities_counter = 0
        next_level_communities = list()
        append_next_level_community = next_level_communities.append
        number_of_communities = 0
        for j in range(reverse_index_csr.shape[1]):
            if j in unavailable_communities:
                continue
            must_break = reverse_index_csr.shape[1] - unavailable_communities_counter
            print(must_break)
            if must_break < 1:
                break
            unavailable_communities[unavailable_communities_counter] = j
            unavailable_communities_counter += 1
            c_j = reverse_index_cols[j]

            indices = community_neighbors(c_j, reverse_index_rows, unavailable_communities, unavailable_communities_counter)

            max_similarity = -1
            community_index = 0
            for jj in indices:
                c_jj = reverse_index_cols[jj]
                similarity = jaccard(c_j, c_jj)
                if similarity > max_similarity:
                    max_similarity = similarity
                    community_index = jj

            jj = community_index
            if max_similarity > 0:
                # Merge two communities
                c_jj = reverse_index_cols[jj]
                c_new = np.union1d(c_j, c_jj)

                flag_1 = np.setdiff1d(c_new, c_j)
                flag_2 = np.setdiff1d(c_new, c_jj)
                if (flag_1.size != 0) and (flag_2.size != 0):
                    for n in c_new:
                        append_level_row(n)
                        append_level_col(number_of_communities)

                    if c_new.size < alpha:
                        append_next_level_community(number_of_communities)

                    number_of_communities += 1
                unavailable_communities[unavailable_communities_counter] = jj
                unavailable_communities_counter += 1

        level_row = np.array(level_row)
        level_col = np.array(level_col)
        level_data = np.ones(level_row.size, dtype=np.float64)
        communities = sparse.coo_matrix((level_data, (level_row, level_col)),
                                        shape=(number_of_nodes, number_of_communities))

        if communities.getnnz() == 0:
            break

        features = sparse.hstack([features, communities])

        reverse_index_csc = copy.copy(communities)
        reverse_index_csc = reverse_index_csc.tocsc()
        reverse_index_csc = reverse_index_csc[:, np.array(next_level_communities)]
        reverse_index_csr = reverse_index_csc.tocsr()

        reverse_index_rows = np.ndarray(number_of_nodes, dtype=np.ndarray)
        reverse_index_cols = np.ndarray(len(next_level_communities), dtype=np.ndarray)
        for n in range(number_of_nodes):
            reverse_index_row = reverse_index_csr.getrow(n)
            reverse_index_rows[n] = reverse_index_row.indices

            if n < len(next_level_communities):
                reverse_index_col = reverse_index_csc.getcol(n)
                reverse_index_cols[n] = reverse_index_col.indices

        if len(next_level_communities) > 1:
            flag = True

        iteration += 1
        print('Iteration: ', iteration)
        print('List length', len(next_level_communities))

    return features


def community_neighbors(c_j, reverse_index_rows, unavailable_communities, unavailable_communities_counter):
    """
    Finds communities with shared nodes to a seed community. Called by mroc.

    Inputs:  - c_j: The seed community for which we want to find which communities overlap.
             - reverse_index_rows: A node to community indicator matrix.
             - unavailable_communities: A set of communities that have already either been merged or failed to merge.
             - unavailable_communities_counter: The number of such communities.

    Outputs: - indices: An array containing the communities that exhibit overlap with the seed community.
    """
    indices = list()
    extend = indices.extend
    for node in c_j:
        extend(reverse_index_rows[node])

    indices = np.array(indices)
    indices = np.setdiff1d(indices, unavailable_communities[:unavailable_communities_counter+1])

    return indices


def jaccard(c_1, c_2):
    """
    Calculates the Jaccard similarity between two sets of nodes. Called by mroc.

    Inputs:  - c_1: Community (set of nodes) 1.
             - c_2: Community (set of nodes) 2.

    Outputs: - jaccard_similarity: The Jaccard similarity of these two communities.
    """
    nom = np.intersect1d(c_1, c_2).size
    denom = np.union1d(c_1, c_2).size
    return nom/denom


def louvain(adjacency_matrix):
    """
    Performs community embedding using the LOUVAIN method.

    Introduced in: Blondel, V. D., Guillaume, J. L., Lambiotte, R., & Lefebvre, E. (2008).
                   Fast unfolding of communities in large networks.
                   Journal of Statistical Mechanics: Theory and Experiment, 2008(10), P10008.

    Inputs:  - A in R^(nxn): Adjacency matrix of an undirected network represented as a SciPy Sparse COOrdinate matrix.

    Outputs: - X in R^(nxC_n): The latent space embedding represented as a SciPy Sparse COOrdinate matrix.
    """
    # Convert to networkx undirected graph.
    adjacency_matrix = nx.from_scipy_sparse_matrix(adjacency_matrix, create_using=nx.Graph())

    # Call LOUVAIN algorithm to calculate a hierarchy of communities.
    tree = community.generate_dendogram(adjacency_matrix, part_init=None)

    # Embed communities
    row = list()
    col = list()
    append_row = row.append
    append_col = col.append

    community_counter = 0
    for i in range(len(tree)):
        partition = community.partition_at_level(tree, i)
        for n, c in partition.items():
            append_row(n)
            append_col(community_counter + c)

        community_counter += max(partition.values()) + 1

    row = np.array(row)
    col = np.array(col)
    data = np.ones(row.size, dtype=np.float64)

    louvain_features = sparse.coo_matrix((data, (row, col)), shape=(len(partition.keys()), community_counter),
                                         dtype=np.float64)

    return louvain_features


def laplacian_eigenmaps(adjacency_matrix, k):
    """
    Performs spectral graph embedding using the graph symmetric normalized Laplacian matrix.

    Introduced in: Belkin, M., & Niyogi, P. (2003).
                   Laplacian eigenmaps for dimensionality reduction and data representation.
                   Neural computation, 15(6), 1373-1396.

    Inputs:  -   A in R^(nxn): Adjacency matrix of an network represented as a SciPy Sparse COOrdinate matrix.
             -              k: The number of eigenvectors to extract.

    Outputs: - X in R^(nxk): The latent space embedding represented as a NumPy array. We discard the first eigenvector.
    """
    # Calculate sparse graph Laplacian.
    laplacian = get_normalized_laplacian(adjacency_matrix)

    # Calculate bottom k+1 eigenvalues and eigenvectors of normalized Laplacian.
    try:
        eigenvalues, eigenvectors = spla.eigsh(laplacian,
                                               k=k,
                                               which='SM',
                                               return_eigenvectors=True)
    except spla.ArpackNoConvergence as e:
        print("ARPACK has not converged.")
        eigenvalue = e.eigenvalues
        eigenvectors = e.eigenvectors

    # Discard the eigenvector corresponding to the zero-valued eigenvalue.
    eigenvectors = eigenvectors[:, 1:]

    return eigenvectors


def replicator_eigenmaps(adjacency_matrix, k):
    """
    Performs spectral graph embedding on the centrality reweighted adjacency matrix

    Inputs:  - A in R^(nxn): Adjacency matrix of an undirected network represented as a scipy.sparse.coo_matrix
             -            k: The number of social dimensions/eigenvectors to extract
             -      max_iter: The maximum number of iterations for the iterative eigensolution method

    Outputs: - S in R^(nxk): The social dimensions represented as a numpy.array matrix
    """
    number_of_nodes = adjacency_matrix.shape[0]

    max_eigenvalue = spla.eigsh(adjacency_matrix,
                                k=1,
                                which='LM',
                                return_eigenvectors=False)

    # Calculate Replicator matrix
    eye_matrix = sparse.eye(number_of_nodes, number_of_nodes, dtype=np.float64)
    eye_matrix = eye_matrix.tocsr()
    eye_matrix.data = eye_matrix.data*max_eigenvalue
    replicator = eye_matrix - adjacency_matrix

    # Calculate bottom k+1 eigenvalues and eigenvectors of normalised Laplacian
    try:
        eigenvalues, eigenvectors = spla.eigsh(replicator,
                                               k=k+1,
                                               which='SM',
                                               return_eigenvectors=True)
    except spla.ArpackNoConvergence as e:
        print("ARPACK has not converged.")
        eigenvalue = e.eigenvalues
        eigenvectors = e.eigenvectors

    eigenvectors = eigenvectors[:, 1:]

    return eigenvectors


def base_communities(adjacency_matrix):
    """
    Forms the community indicator normalized feature matrix for any graph.

    Inputs:  - A in R^(nxn): Adjacency matrix of an undirected network represented as a SciPy Sparse COOrdinate matrix.

    Outputs: - X in R^(nxC_n): The latent space embedding represented as a SciPy Sparse COOrdinate matrix.
    """
    number_of_nodes = adjacency_matrix.shape[0]

    # X = A + I
    adjacency_matrix = adjacency_matrix.tocsr()
    adjacency_matrix = adjacency_matrix.transpose()
    features = sparse.csr_matrix(sparse.eye(number_of_nodes, number_of_nodes)) + adjacency_matrix.tocsr()
    features = features.tocsr()
    features.data = np.ones_like(features.data)

    return features



__author__ = 'Georgios Rizos (georgerizos@iti.gr)'

import os
import math
import numpy as np
import numpy.random as rand

from reveal_graph_embedding.common import get_file_row_generator


def get_folds_generator(node_label_matrix,
                        labelled_node_indices,
                        number_of_categories,
                        dataset_memory_folder,
                        percentage,
                        number_of_folds=10):
    """
    Read or form and store the seed nodes for training and testing.

    Inputs: - node_label_matrix: The node-label ground truth in a SciPy sparse matrix format.
            - labelled_node_indices: A NumPy array containing the labelled node indices.
            - number_of_categories: The number of categories/classes in the learning.
            - memory_path: The folder where the results are stored.
            - percentage: The percentage of labelled samples that will be used for training.

    Output: - folds: A generator containing train/test set folds.
    """
    number_of_labeled_nodes = labelled_node_indices.size
    training_set_size = int(np.ceil(percentage*number_of_labeled_nodes/100))

    ####################################################################################################################
    # Read or generate folds
    ####################################################################################################################
    fold_file_path = dataset_memory_folder + "/folds/" + str(percentage) + "_folds.txt"
    train_list = list()
    test_list = list()
    if not os.path.exists(fold_file_path):
        with open(fold_file_path, "w") as fp:
            for trial in np.arange(number_of_folds):
                train, test = valid_train_test(node_label_matrix[labelled_node_indices, :],
                                               training_set_size,
                                               number_of_categories,
                                               trial)
                train = labelled_node_indices[train]
                test = labelled_node_indices[test]

                # Write test nodes
                row = [str(node) for node in test]
                row = "\t".join(row) + "\n"
                fp.write(row)

                # Write train nodes
                row = [str(node) for node in train]
                row = "\t".join(row) + "\n"
                fp.write(row)

                train_list.append(train)
                test_list.append(test)
    else:
        file_row_gen = get_file_row_generator(fold_file_path, "\t")

        for trial in np.arange(number_of_folds):
            # Read test nodes
            test = next(file_row_gen)
            test = [int(node) for node in test]
            test = np.array(test)

            # Read train nodes
            train = next(file_row_gen)
            train = [int(node) for node in train]
            train = np.array(train)

            train_list.append(train)
            test_list.append(test)

    folds = ((train, test) for train, test in zip(train_list, test_list))
    return folds


def generate_folds(node_label_matrix, labelled_node_indices, number_of_categories, percentage, number_of_folds=10):
    """
    Form the seed nodes for training and testing.

    Inputs:  - node_label_matrix: The node-label ground truth in a SciPy sparse matrix format.
             - labelled_node_indices: A NumPy array containing the labelled node indices.
             - number_of_categories: The number of categories/classes in the learning.
             - percentage: The percentage of labelled samples that will be used for training.

    Output:  - folds: A generator containing train/test set folds.
    """
    number_of_labeled_nodes = labelled_node_indices.size
    training_set_size = int(np.ceil(percentage*number_of_labeled_nodes/100))

    ####################################################################################################################
    # Generate folds
    ####################################################################################################################
    train_list = list()
    test_list = list()
    for trial in np.arange(number_of_folds):
        train, test = valid_train_test(node_label_matrix[labelled_node_indices, :],
                                       training_set_size,
                                       number_of_categories,
                                       trial)
        train = labelled_node_indices[train]
        test = labelled_node_indices[test]

        train_list.append(train)
        test_list.append(test)

    folds = ((train, test) for train, test in zip(train_list, test_list))
    return folds


def valid_train_test(node_label_matrix, training_set_size, number_of_categories, random_seed=0):
    """
    Partitions the labelled node set into training and testing set, making sure one category exists in both sets.

    Inputs:  - node_label_matrix: The node-label ground truth in a SciPy sparse matrix format.
             - training_set_size: The minimum required size for the training set.
             - number_of_categories: The number of categories/classes in the learning.
             - random_seed: A seed for numpy random.

    Outputs: - train_set: A NumPy array containing the training set node ids.
             - test_set: A NumPy array containing the testing set node ids.

    TODO: This function might benefit from some cleaning.
    """
    number_of_labelled_nodes = node_label_matrix.shape[0]

    # Randomize process
    np.random.seed(random_seed)
    perm = rand.permutation(number_of_labelled_nodes)
    node_label_matrix = node_label_matrix[perm, :]

    remaining_nodes = set(list(np.arange(number_of_labelled_nodes)))

    # Choose at least one user for any category for the training set
    train_ids = list()
    for c in np.arange(number_of_categories):
        not_found = True
        for t in remaining_nodes:
            indices = node_label_matrix.getrow(t).indices
            if c in indices:
                train_ids.append(t)
                not_found = False
                break
        if not_found:
            # This should never be reached
            print("Not found enough training data for training set.")
            raise RuntimeError

    # This is done to remove duplicates via the set() structure
    train_ids = set(train_ids)

    remaining_nodes.difference_update(train_ids)

    # This is done for speed
    train_ids = np.array(list(train_ids))

    # Choose at least one user for any category for the testing set
    test_ids = list()
    for c in np.arange(number_of_categories):
        not_found = True
        for t in remaining_nodes:
            indices = node_label_matrix.getrow(t).indices
            if c in indices:
                test_ids.append(t)
                not_found = False
                break
        if not_found:
            # This should never be reached
            print("Not found enough testing data for testing set.")
            raise RuntimeError

    # This is done to remove duplicates via the set() structure
    test_ids = set(test_ids)

    remaining_nodes.difference_update(test_ids)

    # Meet the training set size quota by adding new nodes
    if train_ids.size < training_set_size:
        # Calculate how many more nodes are needed
        remainder = training_set_size - train_ids.size

        # Find the nodes not currently in the training set
        more_train_ids = np.array(list(remaining_nodes))

        # Choose randomly among the nodes
        perm2 = rand.permutation(more_train_ids.size)
        more_train_ids = list(more_train_ids[perm2[:remainder]])
        train_ids = np.array(list(set(list(train_ids) + more_train_ids)))

    remaining_nodes.difference_update(set(list(train_ids)))

    # Form the test set
    test_ids.update(remaining_nodes)
    test_ids = np.array(list(set(list(test_ids))))

    return perm[train_ids], perm[test_ids]


def iterative_stratification(node_label_matrix, training_set_size, number_of_categories, random_seed=0):
    """
    Iterative data fold stratification/balancing for two folds.

    Based on: Sechidis, K., Tsoumakas, G., & Vlahavas, I. (2011).
              On the stratification of multi-label data.
              In Machine Learning and Knowledge Discovery in Databases (pp. 145-158).
              Springer Berlin Heidelberg.

    Inputs:  - node_label_matrix: The node-label ground truth in a SciPy sparse matrix format.
             - training_set_size: The minimum required size for the training set.
             - number_of_categories: The number of categories/classes in the learning.
             - random_seed: A seed for numpy random.

    Outputs: - train_set: A NumPy array containing the training set node ids.
             - test_set: A NumPy array containing the testing set node ids.
    """
    number_of_labelled_nodes = node_label_matrix.shape[0]
    testing_set_size = number_of_labelled_nodes - training_set_size
    training_set_proportion = training_set_size/number_of_labelled_nodes
    testing_set_proportion = testing_set_size/number_of_labelled_nodes

    # Calculate the desired number of examples of each label at each subset.
    desired_label_number = np.zeros((2, number_of_categories), dtype=np.int64)
    node_label_matrix = node_label_matrix.tocsc()
    for j in range(number_of_categories):
        category_label_number = node_label_matrix.getcol(j).indices.size
        desired_label_number[0, j] = math.ceil(category_label_number*training_set_proportion)
        desired_label_number[1, j] = category_label_number - desired_label_number[0, j]

    train_ids = list()
    test_ids = list()

    append_train_id = train_ids.append
    append_test_id = test_ids.append

    # Randomize process
    np.random.seed(random_seed)
    while True:
        if len(train_ids) + len(test_ids) >= number_of_labelled_nodes:
            break
        # Find the label with the fewest (but at least one) remaining examples, breaking the ties randomly
        remaining_label_distribution = desired_label_number.sum(axis=0)
        min_label = np.min(remaining_label_distribution[np.where(remaining_label_distribution > 0)[0]])
        label_indices = np.where(remaining_label_distribution == min_label)[0]
        chosen_label = int(np.random.choice(label_indices, 1)[0])

        # Find the subset with the largest number of desired examples for this label,
        # breaking ties by considering the largest number of desired examples, breaking further ties randomly.
        fold_max_remaining_labels = np.max(desired_label_number[:, chosen_label])
        fold_indices = np.where(desired_label_number[:, chosen_label] == fold_max_remaining_labels)[0]
        chosen_fold = int(np.random.choice(fold_indices, 1)[0])

        # Choose a random example for the selected label.
        relevant_nodes = node_label_matrix.getcol(chosen_label).indices
        chosen_node = int(np.random.choice(np.setdiff1d(relevant_nodes,
                                                        np.union1d(np.array(train_ids),
                                                                   np.array(test_ids))),
                                           1)[0])
        if chosen_fold == 0:
            append_train_id(chosen_node)
            desired_label_number[0, node_label_matrix.getrow(chosen_node).indices] -= 1
        elif chosen_fold == 1:
            append_test_id(chosen_node)
            desired_label_number[1, node_label_matrix.getrow(chosen_node).indices] -= 1
        else:
            raise RuntimeError

    return np.array(train_ids), np.array(test_ids)



__author__ = 'Georgios Rizos (georgerizos@iti.gr)'

import numpy as np
import scipy.sparse as spsp
from collections import defaultdict

from reveal_graph_embedding.common import get_file_row_generator


def read_adjacency_matrix(file_path, separator, numbering="matlab"):
    """
    Reads an edge list in csv format and returns the adjacency matrix in SciPy Sparse COOrdinate format.

    Inputs:  - file_path: The path where the adjacency matrix is stored.
             - separator: The delimiter among values (e.g. ",", "\t", " ")
             - numbering: Array numbering style: * "matlab"
                                                 * "c"

    Outputs: - adjacency_matrix: The adjacency matrix in SciPy Sparse COOrdinate format.
    """
    # Open file
    file_row_generator = get_file_row_generator(file_path, separator)

    file_row = next(file_row_generator)
    number_of_rows = file_row[1]
    number_of_columns = file_row[3]
    directed = file_row[7]
    if directed == "True":
        directed = True
    elif directed == "False":
        directed = False
    else:
        print("Invalid metadata.")
        raise RuntimeError

    # Initialize lists for row and column sparse matrix arguments
    row = list()
    col = list()
    data = list()
    append_row = row.append
    append_col = col.append
    append_data = data.append

    # Read all file rows
    for file_row in file_row_generator:
        source_node = np.int64(file_row[0])
        target_node = np.int64(file_row[1])
        edge_weight = np.float64(file_row[2])

        # Add edge
        append_row(source_node)
        append_col(target_node)
        append_data(edge_weight)

        # Since this is an undirected network also add the reciprocal edge
        if not directed:
            if source_node != target_node:
                append_row(target_node)
                append_col(source_node)
                append_data(edge_weight)

    row = np.array(row, dtype=np.int64)
    col = np.array(col, dtype=np.int64)
    data = np.array(data, dtype=np.float64)

    if numbering == "matlab":
        row -= 1
        col -= 1
    elif numbering == "c":
        pass
    else:
        print("Invalid numbering style.")
        raise RuntimeError

    # Form sparse adjacency matrix
    adjacency_matrix = spsp.coo_matrix((data, (row, col)), shape=(number_of_rows, number_of_columns))

    return adjacency_matrix


def read_node_label_matrix(file_path, separator, numbering="matlab"):
    """
    Reads node-label pairs in csv format and returns a list of tuples and a node-label matrix.

    Inputs:  - file_path: The path where the node-label matrix is stored.
             - separator: The delimiter among values (e.g. ",", "\t", " ")
             - number_of_nodes: The number of nodes of the full graph. It is possible that not all nodes are labelled.
             - numbering: Array numbering style: * "matlab"
                                                 * "c"

    Outputs: - node_label_matrix: The node-label associations in a NumPy array of tuples format.
             - number_of_categories: The number of categories/classes the nodes may belong to.
             - labelled_node_indices: A NumPy array containing the labelled node indices.
    """
    # Open file
    file_row_generator = get_file_row_generator(file_path, separator)

    file_row = next(file_row_generator)
    number_of_rows = file_row[1]
    number_of_categories = int(file_row[3])

    # Initialize lists for row and column sparse matrix arguments
    row = list()
    col = list()
    append_row = row.append
    append_col = col.append

    # Populate the arrays
    for file_row in file_row_generator:
        node = np.int64(file_row[0])
        label = np.int64(file_row[1])

        # Add label
        append_row(node)
        append_col(label)

    labelled_node_indices = np.array(list(set(row)))

    row = np.array(row, dtype=np.int64)
    col = np.array(col, dtype=np.int64)
    data = np.ones_like(row, dtype=np.float64)

    if numbering == "matlab":
        row -= 1
        col -= 1
        labelled_node_indices -= 1
    elif numbering == "c":
        pass
    else:
        print("Invalid numbering style.")
        raise RuntimeError

    # Form sparse adjacency matrix
    node_label_matrix = spsp.coo_matrix((data, (row, col)), shape=(number_of_rows, number_of_categories))
    node_label_matrix = node_label_matrix.tocsr()

    return node_label_matrix, number_of_categories, labelled_node_indices


def scipy_sparse_to_csv(filepath, matrix, separator=",", directed=False, numbering="matlab"):
    """
    Writes sparse matrix in separated value format.
    """
    matrix = spsp.coo_matrix(matrix)

    shape = matrix.shape
    nnz = matrix.getnnz()

    if numbering == "matlab":
        row = matrix.row + 1
        col = matrix.col + 1
        data = matrix.data
    elif numbering == "c":
        row = matrix.row
        col = matrix.col
        data = matrix.data
    else:
        print("Invalid numbering style.")
        raise RuntimeError

    with open(filepath, "w") as f:
        # Write metadata.
        file_row = "n_rows:" + separator + str(shape[0]) + separator +\
                   "n_cols:" + separator + str(shape[1]) + separator +\
                   "nnz:" + separator + str(nnz) + separator +\
                   "directed:" + separator + str(directed) +\
                   "\n"
        f.write(file_row)

        for edge in range(row.size):
            if directed is False:
                if col[edge] < row[edge]:
                    continue
            file_row = str(row[edge]) + separator + str(col[edge]) + separator + str(data[edge]) + "\n"
            f.write(file_row)


def write_screen_name_to_topics(filepath, user_label_matrix, node_to_id, id_to_name, label_to_lemma, lemma_to_keyword, separator=","):
    """
    Writes a user name and associated topic names per row.
    """
    user_label_matrix = spsp.coo_matrix(user_label_matrix)

    shape = user_label_matrix.shape
    nnz = user_label_matrix.getnnz()

    row = user_label_matrix.row
    col = user_label_matrix.col
    data = user_label_matrix.data

    name_to_topic_set = defaultdict(set)

    for edge in range(row.size):
        node = row[edge]
        user_twitter_id = node_to_id[node]
        name = id_to_name[user_twitter_id]

        label = col[edge]
        lemma = label_to_lemma[label]
        # topic = lemma_to_keyword[lemma]

        name_to_topic_set[name].add(lemma)

    with open(filepath, "w") as f:
        # Write metadata.
        file_row = "n_rows:" + separator + str(shape[0]) + separator +\
                   "nnz:" + separator + str(nnz) + separator +\
                   "\n"
        f.write(file_row)

        for name, topic_set in name_to_topic_set.items():
            file_row = list()
            file_row.append(name)
            file_row.extend(topic_set)
            file_row = separator.join(file_row) + "\n"
            f.write(file_row)



__author__ = 'Georgios Rizos (georgerizos@iti.gr)'

from setuptools import setup
from setuptools.extension import Extension
try:
    from Cython.Distutils import build_ext
except ImportError:
    USE_CYTHON = False
else:
    USE_CYTHON = True
import numpy
C_OPT_FLAG = "-O3"


def readme():
    with open("README.md") as f:
        return f.read()

cmdclass = dict()
ext_modules = list()


# Either use cython to translate the .pyx files, or compile the distributed .c files as translated by the author.
if USE_CYTHON:
    ext_modules.append(Extension(name="reveal_graph_embedding.embedding.arcte.cython_opt.arcte",
                                 sources=["reveal_graph_embedding/embedding/arcte/cython_opt/arcte.pyx"],
                                 extra_compile_args=[C_OPT_FLAG, '-I/user/local/include/python3.4'],
                                 include_dirs=[numpy.get_include()]))
    ext_modules.append(Extension(name="reveal_graph_embedding.eps_randomwalk.cython_opt.push",
                                 sources=["reveal_graph_embedding/eps_randomwalk/cython_opt/push.pyx"],
                                 extra_compile_args=[C_OPT_FLAG, '-fopenmp', '-I/user/local/include/python3.4'],
                                 extra_link_args=['-fopenmp'],
                                 include_dirs=[numpy.get_include()]))
    ext_modules.append(Extension(name="reveal_graph_embedding.eps_randomwalk.cython_opt.transition",
                                 sources=["reveal_graph_embedding/eps_randomwalk/cython_opt/transition.pyx"],
                                 extra_compile_args=[C_OPT_FLAG, '-fopenmp', '-I/user/local/include/python3.4'],
                                 extra_link_args=['-fopenmp'],
                                 include_dirs=[numpy.get_include()]))
    ext_modules.append(Extension(name="reveal_graph_embedding.eps_randomwalk.cython_opt.similarity",
                                 sources=["reveal_graph_embedding/eps_randomwalk/cython_opt/similarity.pyx"],
                                 extra_compile_args=[C_OPT_FLAG, '-fopenmp', '-I/user/local/include/python3.4'],
                                 extra_link_args=['-fopenmp'],
                                 include_dirs=[numpy.get_include()]))
    cmdclass.update({"build_ext": build_ext})
else:
    ext_modules.append(Extension(name="reveal_graph_embedding.embedding.arcte.cython_opt.arcte",
                                 sources=["reveal_graph_embedding/embedding/arcte/cython_opt/arcte.c"],
                                 extra_compile_args=[C_OPT_FLAG, '-I/user/local/include/python3.4'],
                                 include_dirs=[numpy.get_include()]))
    ext_modules.append(Extension(name="reveal_graph_embedding.eps_randomwalk.cython_opt.push",
                                 sources=["reveal_graph_embedding/eps_randomwalk/cython_opt/push.c"],
                                 extra_compile_args=[C_OPT_FLAG, '-fopenmp', '-I/user/local/include/python3.4'],
                                 extra_link_args=['-fopenmp'],
                                 include_dirs=[numpy.get_include()]))
    ext_modules.append(Extension(name="reveal_graph_embedding.eps_randomwalk.cython_opt.transition",
                                 sources=["reveal_graph_embedding/eps_randomwalk/cython_opt/transition.c"],
                                 extra_compile_args=[C_OPT_FLAG, '-fopenmp', '-I/user/local/include/python3.4'],
                                 extra_link_args=['-fopenmp'],
                                 include_dirs=[numpy.get_include()]))
    ext_modules.append(Extension(name="reveal_graph_embedding.eps_randomwalk.cython_opt.similarity",
                                 sources=["reveal_graph_embedding/eps_randomwalk/cython_opt/similarity.c"],
                                 extra_compile_args=[C_OPT_FLAG, '-fopenmp', '-I/user/local/include/python3.4'],
                                 extra_link_args=['-fopenmp'],
                                 include_dirs=[numpy.get_include()]))


setup(
    name='reveal-graph-embedding',
    version='0.1.4',
    author='Georgios Rizos',
    author_email='georgerizos@iti.gr',
    packages=['reveal_graph_embedding',
              'reveal_graph_embedding.datautil',
              'reveal_graph_embedding.datautil.asu_datautil',
              'reveal_graph_embedding.datautil.insight_datautil',
              'reveal_graph_embedding.datautil.snow_datautil',
              'reveal_graph_embedding.embedding',
              'reveal_graph_embedding.embedding.arcte',
              'reveal_graph_embedding.embedding.arcte.cython_opt',
              'reveal_graph_embedding.eps_randomwalk',
              'reveal_graph_embedding.eps_randomwalk.benchmarks',
              'reveal_graph_embedding.eps_randomwalk.benchmarks.time_comparisons',
              'reveal_graph_embedding.eps_randomwalk.cython_opt',
              'reveal_graph_embedding.learning',
              'reveal_graph_embedding.quality'],
    cmdclass=cmdclass,
    ext_modules=ext_modules,
    url='https://github.com/MKLab-ITI/reveal-graph-embedding',
    license='Apache',
    description='Implementation of community-based graph embedding for user classification.',
    long_description=readme(),
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Operating System :: Unix',
        'Programming Language :: Cython',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: Implementation',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Scientific/Engineering :: Information Analysis'
    ],
    keywords="online-social-network user-centric-pagerank user-centric-community-detection graph-embedding Reveal-FP7",
    entry_points={
        'console_scripts': ['arcte=reveal_graph_embedding.entry_points.arcte:main'],
    },
    include_package_data=False,
    install_requires=[]
)



__author__ = 'Georgios Rizos (georgerizos@iti.gr)'

import h5py
import numpy as np
import os


def write_predictions(path, data):
    with h5py.File(path + ".h5", "w") as f:
        f.create_dataset('data', data=data, compression="gzip", compression_opts=9)


def read_predictions(path):
    with h5py.File(path + ".h5") as f:
        data = f["/data"][:, :]

        return data


def npy_to_h5(path):
    data = np.load(path + ".npy")

    write_predictions(path, data)

    os.remove(path + ".npy")
